package com.sun.jndi.toolkit.dir;

import java.util.NoSuchElementException;
import java.util.Enumeration;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;
import javax.naming.directory.AttributeModificationException;
import javax.naming.directory.ModificationItem;
import javax.naming.directory.Attribute;
import javax.naming.CompositeName;
import javax.naming.Context;
import javax.naming.Binding;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NameAlreadyBoundException;
import javax.naming.directory.SchemaViolationException;
import javax.naming.spi.DirStateFactory;
import javax.naming.InvalidNameException;
import javax.naming.NameNotFoundException;
import javax.naming.spi.DirectoryManager;
import javax.naming.directory.BasicAttributes;
import javax.naming.OperationNotSupportedException;
import javax.naming.NamingException;
import javax.naming.directory.Attributes;
import javax.naming.Name;
import java.util.Hashtable;
import javax.naming.NameParser;
import javax.naming.directory.DirContext;

public class HierMemDirCtx implements DirContext
{
    private static final boolean debug = false;
    private static final NameParser defaultParser;
    protected Hashtable<String, Object> myEnv;
    protected Hashtable<Name, Object> bindings;
    protected Attributes attrs;
    protected boolean ignoreCase;
    protected NamingException readOnlyEx;
    protected NameParser myParser;
    private boolean alwaysUseFactory;
    
    @Override
    public void close() throws NamingException {
        this.myEnv = null;
        this.bindings = null;
        this.attrs = null;
    }
    
    @Override
    public String getNameInNamespace() throws NamingException {
        throw new OperationNotSupportedException("Cannot determine full name");
    }
    
    public HierMemDirCtx() {
        this(null, false, false);
    }
    
    public HierMemDirCtx(boolean b) {
        this(null, b, false);
    }
    
    public HierMemDirCtx(Hashtable<String, Object> hashtable, boolean b) {
        this(hashtable, b, false);
    }
    
    protected HierMemDirCtx(Hashtable<String, Object> myEnv, boolean ignoreCase, boolean alwaysUseFactory) {
        this.ignoreCase = false;
        this.readOnlyEx = null;
        this.myParser = HierMemDirCtx.defaultParser;
        this.myEnv = myEnv;
        this.ignoreCase = ignoreCase;
        this.init();
        this.alwaysUseFactory = alwaysUseFactory;
    }
    
    private void init() {
        this.attrs = new BasicAttributes(this.ignoreCase);
        this.bindings = new Hashtable(11, 0.75f);
    }
    
    @Override
    public Object lookup(String s) throws NamingException {
        return this.lookup(this.myParser.parse(s));
    }
    
    @Override
    public Object lookup(Name name) throws NamingException {
        return this.doLookup(name, this.alwaysUseFactory);
    }
    
    public Object doLookup(Name canonizeName, boolean b) throws NamingException {
        canonizeName = this.canonizeName(canonizeName);
        HierMemDirCtx refInfo;
        switch (canonizeName.size()) {
            case 0: {
                refInfo = this;
                break;
            }
            case 1: {
                refInfo = this.bindings.get(canonizeName);
                break;
            }
            default: {
                HierMemDirCtx hierMemDirCtx = (HierMemDirCtx)this.bindings.get(canonizeName.getPrefix(1));
                if (hierMemDirCtx == null) {
                    refInfo = null;
                    break;
                }
                refInfo = hierMemDirCtx.doLookup(canonizeName.getSuffix(1), false);
                break;
            }
        }
        if (refInfo == null) {
            throw new NameNotFoundException(canonizeName.toString());
        }
        if (b) {
            try {
                return DirectoryManager.getObjectInstance(refInfo, canonizeName, this, this.myEnv, (refInfo instanceof HierMemDirCtx) ? ((HierMemDirCtx)refInfo).attrs : null);
            }
            catch (NamingException ex) {
                throw ex;
            }
            catch (Exception rootCause) {
                NamingException o = new NamingException("Problem calling getObjectInstance");
                o.setRootCause(rootCause);
                throw o;
            }
        }
        else {
            return refInfo;
        }
    }
    
    @Override
    public void bind(String s, Object o) throws NamingException {
        this.bind(this.myParser.parse(s), o);
    }
    
    @Override
    public void bind(Name name, Object o) throws NamingException {
        this.doBind(name, o, null, this.alwaysUseFactory);
    }
    
    @Override
    public void bind(String s, Object o, Attributes attributes) throws NamingException {
        this.bind(this.myParser.parse(s), o, attributes);
    }
    
    @Override
    public void bind(Name name, Object o, Attributes attributes) throws NamingException {
        this.doBind(name, o, attributes, this.alwaysUseFactory);
    }
    
    protected void doBind(Name name, Object object, Attributes attributes, boolean b) throws NamingException {
        if (name.isEmpty()) {
            throw new InvalidNameException("Cannot bind empty name");
        }
        if (b) {
            DirStateFactory.Result stateToBind = DirectoryManager.getStateToBind(object, name, this, this.myEnv, attributes);
            object = stateToBind.getObject();
            attributes = stateToBind.getAttributes();
        }
        ((HierMemDirCtx)this.doLookup(this.getInternalName(name), false)).doBindAux(this.getLeafName(name), object);
        if (attributes != null && attributes.size() > 0) {
            this.modifyAttributes(name, 1, attributes);
        }
        return;
    }
    
    protected void doBindAux(Name name, Object value) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        if (this.bindings.get(name) != null) {
            throw new NameAlreadyBoundException(name.toString());
        }
        if ((value instanceof HierMemDirCtx)) {
            this.bindings.put(name, value);
            return;
        }
        throw new SchemaViolationException("This context only supports binding objects of it's own kind");
    }
    
    @Override
    public void rebind(String s, Object o) throws NamingException {
        this.rebind(this.myParser.parse(s), o);
    }
    
    @Override
    public void rebind(Name name, Object o) throws NamingException {
        this.doRebind(name, o, null, this.alwaysUseFactory);
    }
    
    @Override
    public void rebind(String s, Object o, Attributes attributes) throws NamingException {
        this.rebind(this.myParser.parse(s), o, attributes);
    }
    
    @Override
    public void rebind(Name name, Object o, Attributes attributes) throws NamingException {
        this.doRebind(name, o, attributes, this.alwaysUseFactory);
    }
    
    protected void doRebind(Name name, Object object, Attributes attributes, boolean b) throws NamingException {
        if (name.isEmpty()) {
            throw new InvalidNameException("Cannot rebind empty name");
        }
        if (b) {
            DirStateFactory.Result stateToBind = DirectoryManager.getStateToBind(object, name, this, this.myEnv, attributes);
            object = stateToBind.getObject();
            attributes = stateToBind.getAttributes();
        }
        ((HierMemDirCtx)this.doLookup(this.getInternalName(name), false)).doRebindAux(this.getLeafName(name), object);
        if (attributes != null && attributes.size() > 0) {
            this.modifyAttributes(name, 1, attributes);
        }
        return;
    }
    
    protected void doRebindAux(Name key, Object value) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        if ((value instanceof HierMemDirCtx)) {
            this.bindings.put(key, value);
            return;
        }
        throw new SchemaViolationException("This context only supports binding objects of it's own kind");
    }
    
    @Override
    public void unbind(String s) throws NamingException {
        this.unbind(this.myParser.parse(s));
    }
    
    @Override
    public void unbind(Name name) throws NamingException {
        if (name.isEmpty()) {
            throw new InvalidNameException("Cannot unbind empty name");
        }
        ((HierMemDirCtx)this.doLookup(this.getInternalName(name), false)).doUnbind(this.getLeafName(name));
        return;
    }
    
    protected void doUnbind(Name key) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        this.bindings.remove(key);
        return;
    }
    
    @Override
    public void rename(String s, String s2) throws NamingException {
        this.rename(this.myParser.parse(s), this.myParser.parse(s2));
    }
    
    @Override
    public void rename(Name name, Name name2) throws NamingException {
        if (name2.isEmpty() || name.isEmpty()) {
            throw new InvalidNameException("Cannot rename empty name");
        }
        if (!this.getInternalName(name2).equals(this.getInternalName(name))) {
            throw new InvalidNameException("Cannot rename across contexts");
        }
        ((HierMemDirCtx)this.doLookup(this.getInternalName(name2), false)).doRename(this.getLeafName(name), this.getLeafName(name2));
        return;
    }
    
    protected void doRename(Name canonizeName, Name canonizeName2) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        canonizeName = this.canonizeName(canonizeName);
        canonizeName2 = this.canonizeName(canonizeName2);
        if (this.bindings.get(canonizeName2) != null) {
            throw new NameAlreadyBoundException(canonizeName2.toString());
        }
        Object remove = this.bindings.remove(canonizeName);
        if (remove == null) {
            throw new NameNotFoundException(canonizeName.toString());
        }
        this.bindings.put(canonizeName2, remove);
        return;
    }
    
    @Override
    public NamingEnumeration<NameClassPair> list(String s) throws NamingException {
        return this.list(this.myParser.parse(s));
    }
    
    @Override
    public NamingEnumeration<NameClassPair> list(Name name) throws NamingException {
        return ((HierMemDirCtx)this.doLookup(name, false)).doList();
    }
    
    protected NamingEnumeration<NameClassPair> doList() throws NamingException {
        return new HierMemDirCtx.FlatNames(this.bindings.keys());
    }
    
    @Override
    public NamingEnumeration<Binding> listBindings(String s) throws NamingException {
        return this.listBindings(this.myParser.parse(s));
    }
    
    @Override
    public NamingEnumeration<Binding> listBindings(Name name) throws NamingException {
        return ((HierMemDirCtx)this.doLookup(name, false)).doListBindings(this.alwaysUseFactory);
    }
    
    protected NamingEnumeration<Binding> doListBindings(boolean b) throws NamingException {
        return new HierMemDirCtx.FlatBindings(this.bindings, this.myEnv, b);
    }
    
    @Override
    public void destroySubcontext(String s) throws NamingException {
        this.destroySubcontext(this.myParser.parse(s));
    }
    
    @Override
    public void destroySubcontext(Name name) throws NamingException {
        ((HierMemDirCtx)this.doLookup(this.getInternalName(name), false)).doDestroySubcontext(this.getLeafName(name));
    }
    
    protected void doDestroySubcontext(Name canonizeName) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        canonizeName = this.canonizeName(canonizeName);
        this.bindings.remove(canonizeName);
        return;
    }
    
    @Override
    public Context createSubcontext(String s) throws NamingException {
        return this.createSubcontext(this.myParser.parse(s));
    }
    
    @Override
    public Context createSubcontext(Name name) throws NamingException {
        return this.createSubcontext(name, null);
    }
    
    @Override
    public DirContext createSubcontext(String s, Attributes attributes) throws NamingException {
        return this.createSubcontext(this.myParser.parse(s), attributes);
    }
    
    @Override
    public DirContext createSubcontext(Name name, Attributes attributes) throws NamingException {
        return ((HierMemDirCtx)this.doLookup(this.getInternalName(name), false)).doCreateSubcontext(this.getLeafName(name), attributes);
    }
    
    protected DirContext doCreateSubcontext(Name canonizeName, Attributes attributes) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        canonizeName = this.canonizeName(canonizeName);
        if (this.bindings.get(canonizeName) != null) {
            throw new NameAlreadyBoundException(canonizeName.toString());
        }
        HierMemDirCtx newCtx = this.createNewCtx();
        this.bindings.put(canonizeName, newCtx);
        if (attributes != null) {
            newCtx.modifyAttributes("", 1, attributes);
            return newCtx;
        }
        return newCtx;
    }
    
    @Override
    public Object lookupLink(String s) throws NamingException {
        return this.lookupLink(this.myParser.parse(s));
    }
    
    @Override
    public Object lookupLink(Name name) throws NamingException {
        return this.lookup(name);
    }
    
    @Override
    public NameParser getNameParser(String s) throws NamingException {
        return this.myParser;
    }
    
    @Override
    public NameParser getNameParser(Name name) throws NamingException {
        return this.myParser;
    }
    
    @Override
    public String composeName(String n, String n2) throws NamingException {
        return this.composeName(new CompositeName(n), new CompositeName(n2)).toString();
    }
    
    @Override
    public Name composeName(Name canonizeName, Name canonizeName2) throws NamingException {
        canonizeName = this.canonizeName(canonizeName);
        canonizeName2 = this.canonizeName(canonizeName2);
        Name name = (Name)(Name)canonizeName2.clone();
        name.addAll(canonizeName);
        return name;
    }
    
    @Override
    public Object addToEnvironment(String key, Object value) throws NamingException {
        this.myEnv = ((this.myEnv == null) ? new Hashtable(11, 0.75f) : ((Hashtable)this.myEnv.clone()));
        return this.myEnv.put(key, value);
    }
    
    @Override
    public Object removeFromEnvironment(String key) throws NamingException {
        if (this.myEnv == null) {
            return null;
        }
        this.myEnv = (Hashtable)this.myEnv.clone();
        return this.myEnv.remove(key);
    }
    
    @Override
    public Hashtable<String, Object> getEnvironment() throws NamingException {
        if (this.myEnv == null) {
            return new Hashtable(5, 0.75f);
        }
        return (Hashtable)this.myEnv.clone();
    }
    
    @Override
    public Attributes getAttributes(String s) throws NamingException {
        return this.getAttributes(this.myParser.parse(s));
    }
    
    @Override
    public Attributes getAttributes(Name name) throws NamingException {
        return ((HierMemDirCtx)this.doLookup(name, false)).doGetAttributes();
    }
    
    protected Attributes doGetAttributes() throws NamingException {
        return (Attributes)this.attrs.clone();
    }
    
    @Override
    public Attributes getAttributes(String s, String[] array) throws NamingException {
        return this.getAttributes(this.myParser.parse(s), array);
    }
    
    @Override
    public Attributes getAttributes(Name name, String[] array) throws NamingException {
        return ((HierMemDirCtx)this.doLookup(name, false)).doGetAttributes(array);
    }
    
    protected Attributes doGetAttributes(String[] array) throws NamingException {
        if (array == null) {
            return this.doGetAttributes();
        }
        BasicAttributes basicAttributes = new BasicAttributes(this.ignoreCase);
        for (int i = 0; i < array.length; ++i) {
            Attribute value = this.attrs.get(array[i]);
            if (value != null) {
                basicAttributes.put(value);
            }
        }
        return basicAttributes;
    }
    
    @Override
    public void modifyAttributes(String s, int n, Attributes attributes) throws NamingException {
        this.modifyAttributes(this.myParser.parse(s), n, attributes);
    }
    
    @Override
    public void modifyAttributes(Name name, int mod_op, Attributes attributes) throws NamingException {
        if (attributes == null || attributes.size() == 0) {
            throw new IllegalArgumentException("Cannot modify without an attribute");
        }
        NamingEnumeration<Object> all = attributes.getAll();
        ModificationItem[] array = new ModificationItem[attributes.size()];
        for (int n = 0; n < array.length && all.hasMoreElements(); ++n) {
            array[n] = new ModificationItem(mod_op, (Attribute)all.next());
        }
        this.modifyAttributes(name, array);
        return;
    }
    
    @Override
    public void modifyAttributes(String s, ModificationItem[] array) throws NamingException {
        this.modifyAttributes(this.myParser.parse(s), array);
    }
    
    @Override
    public void modifyAttributes(Name name, ModificationItem[] array) throws NamingException {
        ((HierMemDirCtx)this.doLookup(name, false)).doModifyAttributes(array);
    }
    
    protected void doModifyAttributes(ModificationItem[] array) throws NamingException {
        if (this.readOnlyEx != null) {
            throw (NamingException)this.readOnlyEx.fillInStackTrace();
        }
        applyMods(array, this.attrs);
        return;
    }
    
    protected static Attributes applyMods(ModificationItem[] array, Attributes attributes) throws NamingException {
        for (int i = 0; i < array.length; ++i) {
            ModificationItem modificationItem = array[i];
            Attribute attribute = modificationItem.getAttribute();
            switch (modificationItem.getModificationOp()) {
                case 1: {
                    Attribute value = attributes.get(attribute.getID());
                    if (value == null) {
                        attributes.put((Attribute)attribute.clone());
                        break;
                    }
                    NamingEnumeration<Object> all = attribute.getAll();
                    while (all.hasMore()) {
                        value.add(all.next());
                    }
                    break;
                }
                case 2: {
                    if (attribute.size() == 0) {
                        attributes.remove(attribute.getID());
                        break;
                    }
                    attributes.put((Attribute)attribute.clone());
                    break;
                }
                case 3: {
                    Attribute value2 = attributes.get(attribute.getID());
                    if (value2 != null) {
                        if (attribute.size() == 0) {
                            attributes.remove(attribute.getID());
                            break;
                        }
                        NamingEnumeration<Object> all2 = attribute.getAll();
                        while (all2.hasMore()) {
                            value2.remove(all2.next());
                        }
                        if (value2.size() == 0) {
                            attributes.remove(attribute.getID());
                            break;
                        }
                        break;
                    }
                    else {
                        break;
                    }
                }
                default: {
                    throw new AttributeModificationException("Unknown mod_op");
                }
            }
        }
        return attributes;
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(String s, Attributes attributes) throws NamingException {
        return this.search(s, attributes, (String[])null);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(Name name, Attributes attributes) throws NamingException {
        return this.search(name, attributes, (String[])null);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(String s, Attributes attributes, String[] array) throws NamingException {
        return this.search(this.myParser.parse(s), attributes, array);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(Name name, Attributes attributes, String[] returningAttributes) throws NamingException {
        HierMemDirCtx hierMemDirCtx = (HierMemDirCtx)this.doLookup(name, false);
        SearchControls searchControls = new SearchControls();
        searchControls.setReturningAttributes(returningAttributes);
        return new LazySearchEnumerationImpl(hierMemDirCtx.doListBindings(false), new ContainmentFilter(attributes), searchControls, this, this.myEnv, false);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(Name name, String s, SearchControls searchControls) throws NamingException {
        return new LazySearchEnumerationImpl(new HierMemDirCtx.HierContextEnumerator((DirContext)this.doLookup(name, false), (searchControls != null) ? searchControls.getSearchScope() : 1), new SearchFilter(s), searchControls, this, this.myEnv, this.alwaysUseFactory);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(Name name, String s, Object[] array, SearchControls searchControls) throws NamingException {
        return this.search(name, SearchFilter.format(s, array), searchControls);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(String s, String s2, SearchControls searchControls) throws NamingException {
        return this.search(this.myParser.parse(s), s2, searchControls);
    }
    
    @Override
    public NamingEnumeration<SearchResult> search(String s, String s2, Object[] array, SearchControls searchControls) throws NamingException {
        return this.search(this.myParser.parse(s), s2, array, searchControls);
    }
    
    protected HierMemDirCtx createNewCtx() throws NamingException {
        return new HierMemDirCtx(this.myEnv, this.ignoreCase);
    }
    
    protected Name canonizeName(Name name) throws NamingException {
        if (!(name instanceof HierarchicalName)) {
            name = new HierarchicalName();
            for (int size = name.size(), i = 0; i < size; ++i) {
                name.add(i, name.get(i));
            }
            return name;
        }
        return name;
    }
    
    protected Name getInternalName(Name name) throws NamingException {
        return name.getPrefix(name.size() - 1);
    }
    
    protected Name getLeafName(Name name) throws NamingException {
        return name.getSuffix(name.size() - 1);
    }
    
    @Override
    public DirContext getSchema(String s) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    @Override
    public DirContext getSchema(Name name) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    @Override
    public DirContext getSchemaClassDefinition(String s) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    @Override
    public DirContext getSchemaClassDefinition(Name name) throws NamingException {
        throw new OperationNotSupportedException();
    }
    
    public void setReadOnly(NamingException readOnlyEx) {
        this.readOnlyEx = readOnlyEx;
    }
    
    public void setIgnoreCase(boolean ignoreCase) {
        this.ignoreCase = ignoreCase;
    }
    
    public void setNameParser(NameParser myParser) {
        this.myParser = myParser;
    }
    
    static {
        defaultParser = new HierarchicalNameParser();
    }
    
    private abstract class BaseFlatNames<T> implements NamingEnumeration<T>
    {
        Enumeration<Name> names;
        
        BaseFlatNames(Enumeration<Name> names) {
            this.names = names;
        }
        
        @Override
        public final boolean hasMoreElements() {
            try {
                return this.hasMore();
            }
            catch (NamingException ex) {
                return false;
            }
        }
        
        @Override
        public final boolean hasMore() throws NamingException {
            return this.names.hasMoreElements();
        }
        
        @Override
        public final T nextElement() {
            try {
                return this.next();
            }
            catch (NamingException ex) {
                throw new NoSuchElementException(ex.toString());
            }
        }
        
        @Override
        public abstract T next() throws NamingException;
        
        @Override
        public final void close() {
            this.names = null;
        }
    }
    
    private final class FlatNames extends HierMemDirCtx.BaseFlatNames<NameClassPair>
    {
        FlatNames(Enumeration<Name> enumeration) {
            super(enumeration);
        }
        
        @Override
        public NameClassPair next() throws NamingException {
            Name key = (Name)this.names.nextElement();
            return new NameClassPair(key.toString(), HierMemDirCtx.this.bindings.get(key).getClass().getName());
        }
    }
    
    private final class FlatBindings extends HierMemDirCtx.BaseFlatNames<Binding>
    {
        private Hashtable<Name, Object> bds;
        private Hashtable<String, Object> env;
        private boolean useFactory;
        
        FlatBindings(Hashtable<Name, Object> bds, Hashtable<String, Object> env, boolean useFactory) {
            super(bds.keys());
            this.env = env;
            this.bds = bds;
            this.useFactory = useFactory;
        }
        
        @Override
        public Binding next() throws NamingException {
            Name name = (Name)this.names.nextElement();
            HierMemDirCtx objectInstance = (HierMemDirCtx)this.bds.get(name);
            if (this.useFactory) {
                Attributes attributes = objectInstance.getAttributes("");
                try {
                    objectInstance = DirectoryManager.getObjectInstance(objectInstance, name, HierMemDirCtx.this, this.env, attributes);
                }
                catch (NamingException ex) {
                    throw ex;
                }
                catch (Exception rootCause) {
                    NamingException ex2 = new NamingException("Problem calling getObjectInstance");
                    ex2.setRootCause(rootCause);
                    throw ex2;
                }
            }
            return new Binding(name.toString(), objectInstance);
        }
    }
    
    public class HierContextEnumerator extends ContextEnumerator
    {
        public HierContextEnumerator(Context context, int n) throws NamingException {
            super(context, n);
        }
        
        protected HierContextEnumerator(Context context, int n, String s, boolean b) throws NamingException {
            super(context, n, s, b);
        }
        
        @Override
        protected NamingEnumeration<Binding> getImmediateChildren(Context context) throws NamingException {
            return ((HierMemDirCtx)context).doListBindings(false);
        }
        
        @Override
        protected ContextEnumerator newEnumerator(Context context, int n, String s, boolean b) throws NamingException {
            return new HierMemDirCtx.HierContextEnumerator(context, n, s, b);
        }
    }
}
